/********************************************************
 *  ██████╗  ██████╗████████╗██╗
 * ██╔════╝ ██╔════╝╚══██╔══╝██║
 * ██║  ███╗██║        ██║   ██║
 * ██║   ██║██║        ██║   ██║
 * ╚██████╔╝╚██████╗   ██║   ███████╗
 *  ╚═════╝  ╚═════╝   ╚═╝   ╚══════╝
 * Geophysical Computational Tools & Library (GCTL)
 *
 * Copyright (c) 2023  Yi Zhang (yizhang-geo@zju.edu.cn)
 *
 * GCTL is distributed under a dual licensing scheme. You can redistribute 
 * it and/or modify it under the terms of the GNU Lesser General Public 
 * License as published by the Free Software Foundation, either version 2 
 * of the License, or (at your option) any later version. You should have 
 * received a copy of the GNU Lesser General Public License along with this 
 * program. If not, see <http://www.gnu.org/licenses/>.
 * 
 * If the terms and conditions of the LGPL v.2. would prevent you from using 
 * the GCTL, please consider the option to obtain a commercial license for a 
 * fee. These licenses are offered by the GCTL's original author. As a rule, 
 * licenses are provided "as-is", unlimited in time for a one time fee. Please 
 * send corresponding requests to: yizhang-geo@zju.edu.cn. Please do not forget 
 * to include some description of your company and the realm of its activities. 
 * Also add information on how to contact you by electronic and paper mail.
 ******************************************************/

#ifndef _GCTL_REFELLIPSOID_H
#define _GCTL_REFELLIPSOID_H

#include "../poly/vertex.h"
#include "gmath.h"

namespace gctl
{
    enum refellipsoid_type_e
    {
        // reference system named as the planet name represents a mean radius reference sphere
        Earth,
        MagEarth,
        Moon,
        Mars,
        // Reference ellipsoids 
        WGS84,
        Ardalan2010,
    };

    /**
     * @brief Reference ellipsoid object. Predefined ellipsoids include 
     * Earth, Moon, Mars, WGS84 and Ardalan2010.
     * 
     */
    class refellipsoid
    {
    public:
        refellipsoid(/* args */);
        refellipsoid(refellipsoid_type_e refellipsoid);
        refellipsoid(double R, double r_or_alpha, bool is_alpha = false);
        virtual ~refellipsoid();

        /**
         * @brief Initiate the reference ellipsoid using predefined parameters
         * 
         * @param refellipsoid Predefined reference ellipsoid which includes Earth, Moon, Mars, WGS84 and Ardalan2010
         */
        void set(refellipsoid_type_e refellipsoid);

        /**
         * @brief Initiate the reference ellipsoid using costumed parameters
         * 
         * @param R Equator radius
         * @param r_or_alpha Pole radius or the denominator of the flattening parameter 
         * @param is_flat Is the second paramter indicates the flattening parameter or not
         */
        void set(double R, double r_or_flat, bool is_flat = false);

        /**
         * @brief Compute the local rRdius of curvature on the reference ellipsoid
         * 
         * @param geodetic_lati the inquiring latitude
         * @return radius
         */
        double geodetic_radius(double geodetic_lati);

        /**
         * @brief Converts Geodetic coordinates to Spherical coordinates
         * 
         * @param geodetic_lati Geodetic latitude
         * @param geodetic_hei Height above the ellipsoid
         * @param sph_lati Spherical latitude
         * @param sph_rad Spherical radius
         */
        void geodetic2spherical(double geodetic_lati, double geodetic_hei, 
            double& sph_lati, double& sph_rad);

        void spherical2geodetic(const point3ds& ps, double& geodetic_lon, double& geodetic_lati, 
            double& geodetic_hei, double eps = 1e-16, int cnt = 30);

        /**
         * @brief Convert xyz coordinates to Geodetic coodinates
         * 
         * @param pc A point
         * @param geodetic_lon Geodetic longitude
         * @param geodetic_lati Geodetic latitude
         * @param geodetic_hei Height above the ellipsoid
         * @param eps solving precision
         * @param cnt maximal iteration
         */
        void xyz2geodetic(const point3dc& pc, double& geodetic_lon, double& geodetic_lati, 
            double& geodetic_hei, double eps = 1e-16, int cnt = 30);

        void geodetic2xyz(double geodetic_lon, double geodetic_lati, 
            double geodetic_hei, point3dc& pc);

    private:
        double r_, R_, f_; // semi-minor, semi-major and flat rate
        double e_;
    };
    
};

#endif // _GCTL_REFELLIPSOID_H